package api

import (
	"database/sql"
	"fmt"
	"log"
	"net/http"
	"time"

	"git.hatecomputers.club/hatecomputers/hatecomputers.club/adapters/external_dns/cloudflare"
	"git.hatecomputers.club/hatecomputers/hatecomputers.club/adapters/files/filesystem"
	"git.hatecomputers.club/hatecomputers/hatecomputers.club/api/auth"
	"git.hatecomputers.club/hatecomputers/hatecomputers.club/api/dns"
	"git.hatecomputers.club/hatecomputers/hatecomputers.club/api/guestbook"
	"git.hatecomputers.club/hatecomputers/hatecomputers.club/api/hcaptcha"
	"git.hatecomputers.club/hatecomputers/hatecomputers.club/api/kennel"
	"git.hatecomputers.club/hatecomputers/hatecomputers.club/api/keys"
	"git.hatecomputers.club/hatecomputers/hatecomputers.club/api/profiles"
	"git.hatecomputers.club/hatecomputers/hatecomputers.club/api/template"
	"git.hatecomputers.club/hatecomputers/hatecomputers.club/api/types"
	"git.hatecomputers.club/hatecomputers/hatecomputers.club/args"
	"git.hatecomputers.club/hatecomputers/hatecomputers.club/utils"
)

func LogRequestContinuation(context *types.RequestContext, req *http.Request, resp http.ResponseWriter) types.ContinuationChain {
	return func(success types.Continuation, _failure types.Continuation) types.ContinuationChain {
		context.Start = time.Now()
		context.Id = utils.RandomId()

		log.Println(req.Method, req.URL.Path, req.RemoteAddr, context.Id)
		return success(context, req, resp)
	}
}

func LogExecutionTimeContinuation(context *types.RequestContext, req *http.Request, resp http.ResponseWriter) types.ContinuationChain {
	return func(success types.Continuation, _failure types.Continuation) types.ContinuationChain {
		end := time.Now()
		log.Println(context.Id, "took", end.Sub(context.Start))

		return success(context, req, resp)
	}
}

func HealthCheckContinuation(context *types.RequestContext, req *http.Request, resp http.ResponseWriter) types.ContinuationChain {
	return func(success types.Continuation, _failure types.Continuation) types.ContinuationChain {
		resp.WriteHeader(200)
		resp.Write([]byte("healthy"))
		return success(context, req, resp)
	}
}

func FailurePassingContinuation(context *types.RequestContext, req *http.Request, resp http.ResponseWriter) types.ContinuationChain {
	return func(_success types.Continuation, failure types.Continuation) types.ContinuationChain {
		return failure(context, req, resp)
	}
}

func IdContinuation(context *types.RequestContext, req *http.Request, resp http.ResponseWriter) types.ContinuationChain {
	return func(success types.Continuation, _failure types.Continuation) types.ContinuationChain {
		return success(context, req, resp)
	}
}

func CacheControlMiddleware(next http.Handler, maxAge int) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		header := fmt.Sprintf("public, max-age=%d", maxAge)
		w.Header().Set("Cache-Control", header)
		next.ServeHTTP(w, r)
	})
}

func MakeServer(argv *args.Arguments, dbConn *sql.DB) *http.Server {
	mux := http.NewServeMux()

	// "dependency injection"
	cloudflareAdapter := &cloudflare.CloudflareExternalDNSAdapter{
		APIToken: argv.CloudflareToken,
		ZoneId:   argv.CloudflareZone,
	}
	uploadAdapter := &filesystem.FilesystemAdapter{
		BasePath:    argv.UploadPath,
		Permissions: 0777,
	}

	makeRequestContext := func() *types.RequestContext {
		return &types.RequestContext{
			DBConn:       dbConn,
			Args:         argv,
			TemplateData: &map[string]interface{}{},
		}
	}

	staticFileServer := http.FileServer(http.Dir(argv.StaticPath))
	uploadFileServer := http.FileServer(http.Dir(argv.UploadPath))
	mux.Handle("GET /static/", http.StripPrefix("/static/", CacheControlMiddleware(staticFileServer, 3600)))
	mux.Handle("GET /uploads/", http.StripPrefix("/uploads/", CacheControlMiddleware(uploadFileServer, 60)))

	mux.HandleFunc("GET /", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(IdContinuation, IdContinuation)(auth.ListUsersContinuation, auth.ListUsersContinuation)(template.TemplateContinuation("home.html", true), FailurePassingContinuation)(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("GET /health", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(HealthCheckContinuation, FailurePassingContinuation)(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("GET /login", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.StartSessionContinuation, FailurePassingContinuation)(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("GET /auth", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.InterceptOauthCodeContinuation, FailurePassingContinuation)(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("GET /logout", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.LogoutContinuation, FailurePassingContinuation)(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("GET /profile", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(profiles.GetProfileContinuation, auth.GoLoginContinuation)(template.TemplateContinuation("profile.html", true), template.TemplateContinuation("profile.html", true))(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("POST /profile", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		updateProfileContinuation := profiles.UpdateProfileContinuation(uploadAdapter, profiles.MaxAvatarSize, profiles.AvatarPath, profiles.AvatarPrefix)

		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(updateProfileContinuation, auth.GoLoginContinuation)(profiles.GetProfileContinuation, FailurePassingContinuation)(template.TemplateContinuation("profile.html", true), template.TemplateContinuation("profile.html", true))(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("GET /dns", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(dns.ListDNSRecordsContinuation, auth.GoLoginContinuation)(template.TemplateContinuation("dns.html", true), FailurePassingContinuation)(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("POST /dns", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		createRecordContinuation := dns.CreateDNSRecordContinuation(cloudflareAdapter, dns.MaxUserRecords, dns.UserOwnedInternalFmtDomains)
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(dns.ListDNSRecordsContinuation, auth.GoLoginContinuation)(createRecordContinuation, FailurePassingContinuation)(dns.ListDNSRecordsContinuation, dns.ListDNSRecordsContinuation)(template.TemplateContinuation("dns.html", true), template.TemplateContinuation("dns.html", true))(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("POST /dns/delete", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(dns.DeleteDNSRecordContinuation(cloudflareAdapter), auth.GoLoginContinuation)(dns.ListDNSRecordsContinuation, dns.ListDNSRecordsContinuation)(template.TemplateContinuation("dns.html", true), template.TemplateContinuation("dns.html", true))(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("GET /keys", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(keys.ListAPIKeysContinuation, auth.GoLoginContinuation)(template.TemplateContinuation("api_keys.html", true), FailurePassingContinuation)(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("POST /keys", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(keys.CreateAPIKeyContinuation, auth.GoLoginContinuation)(keys.ListAPIKeysContinuation, keys.ListAPIKeysContinuation)(template.TemplateContinuation("api_keys.html", true), template.TemplateContinuation("api_keys.html", true))(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("POST /keys/delete", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(keys.DeleteAPIKeyContinuation, auth.GoLoginContinuation)(keys.ListAPIKeysContinuation, keys.ListAPIKeysContinuation)(template.TemplateContinuation("api_keys.html", true), template.TemplateContinuation("api_keys.html", true))(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("GET /guestbook", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(hcaptcha.CaptchaArgsContinuation, hcaptcha.CaptchaArgsContinuation)(guestbook.ListGuestbookContinuation, guestbook.ListGuestbookContinuation)(template.TemplateContinuation("guestbook.html", true), FailurePassingContinuation)(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("POST /guestbook", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(hcaptcha.CaptchaVerificationContinuation, hcaptcha.CaptchaVerificationContinuation)(guestbook.SignGuestbookContinuation, FailurePassingContinuation)(guestbook.ListGuestbookContinuation, guestbook.ListGuestbookContinuation)(hcaptcha.CaptchaArgsContinuation, hcaptcha.CaptchaArgsContinuation)(template.TemplateContinuation("guestbook.html", true), template.TemplateContinuation("guestbook.html", true))(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("GET /kennel", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(kennel.GetCatStateAtContinuation, FailurePassingContinuation)(template.TemplateContinuation("kennel_enc.json", false), FailurePassingContinuation)(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("GET /kennel/cats", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(kennel.ListUserCatsContinuation, auth.GoLoginContinuation)(template.TemplateContinuation("kennel_cats.html", true), FailurePassingContinuation)(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("POST /kennel/cats", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		createCatContinuation := kennel.CreateCatContinuation(uploadAdapter, kennel.MaxUserCats, kennel.MaxCatSize, kennel.CatsPath, kennel.CatsPrefix, kennel.DefaultCatSpritesheet)
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(createCatContinuation, FailurePassingContinuation)(kennel.ListUserCatsContinuation, kennel.ListUserCatsContinuation)(template.TemplateContinuation("kennel_cats.html", true), template.TemplateContinuation("kennel_cats.html", true))(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("POST /kennel/cats/delete", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		deleteCatContinuation := kennel.RemoveCatContinuation(uploadAdapter, kennel.CatsPath)

		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(deleteCatContinuation, FailurePassingContinuation)(kennel.ListUserCatsContinuation, FailurePassingContinuation)(template.TemplateContinuation("kennel_cats.html", true), FailurePassingContinuation)(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	mux.HandleFunc("GET /{template}", func(w http.ResponseWriter, r *http.Request) {
		requestContext := makeRequestContext()
		templateFile := r.PathValue("template")
		LogRequestContinuation(requestContext, r, w)(auth.VerifySessionContinuation, FailurePassingContinuation)(IdContinuation, IdContinuation)(template.TemplateContinuation(templateFile+".html", true), FailurePassingContinuation)(LogExecutionTimeContinuation, LogExecutionTimeContinuation)(IdContinuation, IdContinuation)
	})

	return &http.Server{
		Addr:    ":" + fmt.Sprint(argv.Port),
		Handler: mux,
	}
}
